package circuit;

// ---------------------------------------------

import java.util.ArrayList;
import java.util.Iterator;

// ---------------------------------------------

public class CircuitComponent
{
	private String name;
	
	private String type;
	
	protected ArrayList<String> gates, inputs, outputs;		// La lista 'gates'  l'unione di 'inputs' e 'outputs'. 
	
	protected ArrayList<CircuitComponent> components;		// Lista dei componenti costituenti.
	
	protected ArrayList<CircuitLink> links;					// Lista dei collegamenti tra componenti.
	
	
	public CircuitComponent( String name, String type )
	{
		this.name = name;
		
		this.type = type;
		
		gates = new ArrayList();
	
		inputs = new ArrayList();
		
		outputs = new ArrayList();
		
		components = new ArrayList();
		
		links = new ArrayList();
	}
	
	// ------------- METODI DI SETTAGGIO -------------
	// Pu essere utile settare il nome e il tipo di 
	// un componente: ad esempio nel caso in cui si
	// voglia rinominare il componente principale,
	// dandogli nome (e tipo) 'main' (ci  necessario
	// per il funzionamento di NuSMV: deve esservi il
	// modulo main, che utilizza gli altri moduli).
		
	public void setName( String name )
	{
		this.name = name;
	}
		
	public void setType( String type )
	{
		this.type = type;
	}
	
	// -----------------------------------------------
	
	public String getName()
	{
		return name;
	}
	
	public String getType()
	{
		return type;
	}
	
	public int getGateCount()
	{
		return gates.size();
	}
	
	public int getInputCount()
	{
		return inputs.size();
	}
	
	public int getOutputCount()
	{
		return outputs.size();
	}
	
	public int getComponentCount()
	{
		return components.size();
	}
	
	public int getLinkCount()
	{
		return links.size();
	}
	
	public String getGate( int index )
	{
		if ( index >= 0 && index < gates.size() )
			
			return ( String )( gates.get( index ) );
			
		else
			
			return null;
	}
	
	public String getInput( int index )
	{
		if ( index >= 0 && index < inputs.size() )
			
			return ( String )( inputs.get( index ) );
			
		else
			
			return null;
	}		
	
	public boolean isInput( String input )
	{
		return inputs.indexOf( input ) != -1;
	}
	
	public String getOutput( int index )
	{
		if ( index >= 0 && index < outputs.size() )
			
			return ( String )( outputs.get( index ) );
			
		else
			
			return null;
	}	
	
	public boolean isOutput( String output )
	{
		return outputs.indexOf( output ) != -1;
	}
	
	public CircuitComponent getComponent( int index )
	{
		if ( index >= 0 && index < components.size() )
			
			return ( CircuitComponent )( components.get( index ) );
			
		else
			
			return null;
	}
	
	public CircuitLink getLink( int index )
	{
		if ( index >= 0 && index < links.size() )
			
			return links.get( index );
			
		else
			
			return null;
	}
	
	public void addGate( String gate )
	{
		// System.out.println( "Inside component '" + name + "': added gate '" + gate + "'." );
		
		// Convertiamo in lower case perch altrimenti NuSMV pu dare problemi.
		
		gates.add( gate.toLowerCase() );
	}
	
	public void addInput( String input )
	{
		// System.out.println( "Inside component '" + name + "': added input '" + input + "'." );
		
		// Convertiamo in lower case perch altrimenti NuSMV pu dare problemi.
		
		inputs.add( input.toLowerCase() );
	}
	
	public void addOutput( String output )
	{
		// System.out.println( "Inside component '" + name + "': added output '" + output + "'." );
		
		// Convertiamo in lower case perch altrimenti NuSMV pu dare problemi.
		
		outputs.add( output.toLowerCase() );
	}
	
	public void addComponent( CircuitComponent c )
	{
		components.add( c );
	}
	
	public void addLink( CircuitLink l )
	{
		links.add( l );
	}
	
	public void addComponentWire( CircuitComponent c, int componentGateIndex, String wire )
	{	
		wire = wire.toLowerCase();
		
		// In base all'indice 'componentGateIndex', determiniamo se si tratta di un input o di un output. 
		
		boolean isInput = c.inputs.contains( c.gates.get( componentGateIndex ) );
				
		// Dobbiamo effettuare una ricerca nella tabella dei collegamenti, per trovare un eventuale
		// matching. Inizialmente poniamo 'found' a false.		
				
		boolean found = false;
		
		// Se si tratta di un input...
		
		if ( isInput )		
		{
			// Otteniamo il numero di righe della tabella dei collegamenti.
			
			int linkCount = links.size();
			
			// Sfogliamo la tabella dei collegamenti, alla ricerca di un matching.
			
			for ( int i = 0; i < linkCount; i++ )
			{
				// Otteniamo il link ( = quintupla = riga della tabella) corrente.
				
				CircuitLink link = ( CircuitLink )( links.get( i ) );
				
				// Se il nome del filo usato nel link corrente coincide con quello passato come parametro...
				
				if ( wire.equals( link.getWire() ) )
				{
					// ...se nel link corrente era gi presente un componente con il ruolo di input...
					
					if ( link.getComponentIn() != null )
					{
						// ...allora dobbiamo creare un nuovo link con il componente passato come parametro,
						// aggiungendo tale nuovo link alla tabella ( --> nuova riga ). Il nuovo link avr lo stesso
						// componente di output, lo stesso indice di output e lo stesso filo del link trovato come matching.
						
						links.add( new CircuitLink( c, c.inputs.indexOf( c.gates.get( componentGateIndex ) ), link.getComponentOut(), link.getIndexOut(), wire ) );
					
						// Usciamo dal ciclo, per evitare di riaggiungere inutilmente lo stesso
						// collegamento (si tratterebbe di una riga duplicata nella tabella dei
						// collegamenti).
					
						//break;
					}
					// ...altrimenti, se nella riga corrente non era gi presente un componente con il ruolo di input...
					
					else
					{
						// ...allora aggiungiamo al link corrente il componente passato come parametro...
						
						link.setComponentIn( c );
						
						// ...e il relativo indice di input.
						
						link.setIndexIn( c.inputs.indexOf( c.gates.get( componentGateIndex ) ) );
					}					
					
					// Poniamo 'found' a true, poich abbiamo trovato un matching.
						
					found = true;	
				}
			}
			
			// Se sfogliano tutta la tabella non abbiamo trovato alcun matching...
			
			if ( !found )
				
				// ...dobbiamo creare un nuovo link: esso non avr alcun componente di output ( quindi C_out = null e j = -1 ).
				
				links.add( new CircuitLink( c, c.inputs.indexOf( c.gates.get( componentGateIndex ) ), null, -1, wire ) );				
		}	
		
		// ...altrimenti, trattandosi di un output...
			
		else
		{
			int linkCount = links.size();
			
			for ( int i = 0; i < linkCount; i++ )
			{
				CircuitLink link = ( CircuitLink )( links.get( i ) );
				
				if ( wire.equals( link.getWire() ) )
				{
					if ( link.getComponentOut() != null )
					{				
						//System.out.println( "ATTENTION! Short circuit, two different outputs connected together! Previous output will be replaced." );
						
						//System.out.println( "Details: component '" + c.name + "' (at output n " + componentGateIndex + ") is connected to component '" + link.getComponentOut().name + "' (at output n " + link.getIndexOut() + "), using wire '" + wire + "'.\n" );
						
						//link.setComponentOut( c );
					
						//link.setIndexOut( c.outputs.indexOf( c.gates.get( componentGateIndex ) ) );
						
						links.add( new CircuitLink( link.getComponentIn(), link.getIndexIn(), c, c.outputs.indexOf( c.gates.get( componentGateIndex ) ), wire ) );
					}							
					
					else
					{
						link.setComponentOut( c );
						
						link.setIndexOut( c.outputs.indexOf( c.gates.get( componentGateIndex ) ) );
					}					
						
					found = true;	
				}
			}
			
			if ( !found )
				
				links.add( new CircuitLink( null, -1, c, c.outputs.indexOf( c.gates.get( componentGateIndex ) ), wire ) );
		}
	}
	
	public String toString()
	{
		String result = "----- COMPONENT -----\n";
		
		
		result += "\tName:\t" + name + "\n";
		
		
		result += "\tGates:\n";
		
		for ( int i = 0; i < gates.size(); i++ )
			
			result += "\t\t" + gates.get( i ) + "\n";
			
		
		result += "\tInputs:\n";
			
		for ( int i = 0; i < inputs.size(); i++ )
			
			result += "\t\t" + inputs.get( i ) + "\n";
		
		
		result += "\tOutputs:\n";
			
		for ( int i = 0; i < outputs.size(); i++ )
			
			result += "\t\t" + outputs.get( i ) + "\n";
			
		
		result += "\tComponents:\n";
			
		for ( int i = 0; i < components.size(); i++ )
		{
			CircuitComponent c = ( CircuitComponent )( components.get( i ) );
			
			result += "\t\t" + c.name + "( " + c.getInputCount() + " inputs, " + c.getOutputCount() + " outputs)\n";	
		}		
		
		
		result += "\tLinks\n\t\tC_I\t\t\ti\t\tC_O\t\t\tj\t\tw\n";
		
		for ( int i = 0; i < links.size(); i++ )
		{
			CircuitLink l = ( CircuitLink )( links.get( i ) );
					
			result += "\t\t" + ( l.getComponentIn() != null ? l.getComponentIn().name : "<empty>" ) + "\t\t" + l.getIndexIn() + "\t\t" + ( l.getComponentOut() != null ? l.getComponentOut().name : "<empty>" ) + "\t\t" + l.getIndexOut() + "\t\t" + l.getWire() + "\n";	
		}
			
		
		result += "--- END COMPONENT ---\n";
		
		
		return result;	
	}
}