package parser;

// ---------------------------------------------

import circuit.CircuitComponent;
import circuit.CircuitComponent_AND;
import circuit.CircuitComponent_OR;
import circuit.CircuitComponent_NOT;
import circuit.CircuitComponent_NAND;
import circuit.CircuitComponent_NOR;
import circuit.CircuitComponent_XOR;
import circuit.CircuitComponent_XNOR;
import circuit.CircuitComponent_FF_D;
import circuit.CircuitComponent_INV3STATE;

import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

// ---------------------------------------------

public class VerilogParser
{
	private Iterator tokens;
	
	private ArrayList<CircuitComponent> components;
	
	private CircuitComponent currentComponent;
	
	private CircuitComponent mainComponent;
	
	
	public VerilogParser( String fileName )
	{
		String fileContent = readFileContent( fileName );
		
		// System.out.println( "--- FILE TO BE PARSED: ---\n\n" + fileContent + "\n\n--------------------------\n\n" );	
		
		
		components = new ArrayList();
		
		
		tokens = buildTokenIterator( fileContent );
									
		while ( tokens.hasNext() )
		{
			String currentToken = ( String )( tokens.next() );
			
			if ( currentToken.equals( "module" ) )
			
				module( tokens );	
		}
		
		
		for ( int i = 0; i < components.size(); i++ )
		{
			CircuitComponent c = ( CircuitComponent )( components.get( i ) );
			
			// System.out.println( c );
		}
		
		
		mainComponent = getMainComponent();
					
		// System.out.println( "Main component is: " + mainComponent.getName() );	
	
		mainComponent.setName( "main" );
		
		mainComponent.setType( "main" );
	}
	
	public int getComponentCount()
	{
		return components.size();
	}
	
	public CircuitComponent getMainComponent()
	{
		// Insieme contenente i tipi dei vari componenti ( = moduli ) usati nel circuito.
		
		HashSet componentTypes = new HashSet();
		
		// Iniziamo inserendo nell'insieme i tipi dei vari componenti ( componente = modulo nel file Verilog ).
		// NOTA: per i componenti dichiarati come 'module' nel file Verilog, il nostro parser fa coincidere 
		// il loro nome con il loro tipo (essi infatti avrebbero solo il tipo, ma non il nome, poich
		// rappresentano una sorta di livello intensionale). 
		
		for ( int i = 0; i < components.size(); i++ )
		{
			CircuitComponent currentComponent = ( CircuitComponent )( components.get( i ) );
			
			componentTypes.add( currentComponent.getType() ); 
		}	
		
		// Togliamo dall'insieme iniziale ogni tipo di componente che viene utilizzato da altri componenti.
		// Alla fine rimarr solo il tipo del componente principale (main): il componente principale 
		// infatti quello che non viene utilizzato da alcun altro componente.
		
		for ( int i = 0; i < components.size(); i++ )
		{
			CircuitComponent currentComponent = ( CircuitComponent )( components.get( i ) );
			
			for ( int j = 0; j < currentComponent.getComponentCount(); j++ )
			{
				CircuitComponent currentSubComponent = currentComponent.getComponent( j );
			
				componentTypes.remove( currentSubComponent.getType() ); 
			}			
		}
				
		// Avendo a disposizione il tipo del componente principale (il tipo  una stringa), dobbiamo risalire
		// al componente vero e proprio. Ci  possibile senza ambiguit, dal momento che, per i componenti
		// dichiarati come 'module' nel file Verilog, esiste una corrispondenza biunivoca tra essi e i loro tipi.
								
		CircuitComponent mainComponent = null;
		
		String mainComponentType = ( String )( componentTypes.iterator().next() );
		
		for ( int i = 0; i < components.size(); i++ )
		{
			CircuitComponent currentComponent = ( CircuitComponent )( components.get( i ) );
			
			if ( currentComponent.getType().equals( mainComponentType ) )
				
				mainComponent = currentComponent; 
		}
		
		return mainComponent;				
	}
	
	public CircuitComponent getComponent( int index )
	{
		if ( index >= 0 && index < components.size() )
			
			return components.get( index );
			
		else
			
			return null;
	}
	
	private void module( Iterator tokens )
	{
		// System.out.println( "--- MODULE ---" );
				
		
		moduleSignature( tokens );
			
		tokens.next();	// Punto e virgola
			
		moduleBody( tokens );
					
				
		// System.out.println( "--- END MODULE ---\n" );
	}
	
	private void moduleSignature( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		currentComponent = new CircuitComponent( currentToken, currentToken );
		
		components.add( currentComponent );
		
		// System.out.println( "\tName:\t\t\t" + currentToken );
		
		tokens.next();	// Parentesi aperta
		
		currentToken = ( String )( tokens.next() );
					
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.println( "\tParam:\t\t\t" + currentToken );
			
			currentComponent.addGate( currentToken );
			
			currentToken = ( String )( tokens.next() );
		}		
	}
	
	private void moduleBody( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		while ( !currentToken.equals( "endmodule" ) )
		{
			if ( currentToken.equals( "input" ) )
			
				moduleInputs( tokens );
				
			else if ( currentToken.equals( "output" ) )
				
				moduleOutputs( tokens );
				
			else if ( currentToken.equals( "wire" ) )
				
				moduleWires( tokens );
				
			else
			
				moduleComponents( tokens, currentToken );
						
				
			currentToken = ( String )( tokens.next() );	
		}	
	}
	
	private void moduleInputs( Iterator tokens )
	{				
		String currentToken = ( String )( tokens.next() );
		
		while ( !currentToken.equals( ";" ) )
		{
			// System.out.println( "\tInput\t\t\t" + currentToken );
		
			currentComponent.addInput( currentToken );
		
			currentToken = ( String )( tokens.next() );
		}			
	}
	
	private void moduleOutputs( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		while ( !currentToken.equals( ";" ) )
		{
			// System.out.println( "\tOutput\t\t\t" + currentToken );
		
			currentComponent.addOutput( currentToken );
		
			currentToken = ( String )( tokens.next() );
		}	
	}
	
	private void moduleWires( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		while ( !currentToken.equals( ";" ) )
		{
			// System.out.println( "\tWire\t\t\t" + currentToken );
		
			currentToken = ( String )( tokens.next() );
		}			
	}
	
	private void moduleComponents( Iterator tokens, String currentToken )
	{				
		if ( currentToken.equals( "and" ) )
			
			moduleComponent_AND( tokens );
			
		else if ( currentToken.equals( "or" ) )
			
			moduleComponent_OR( tokens );
		
		else if ( currentToken.equals( "not" ) )
			
			moduleComponent_NOT( tokens );
			
		else if ( currentToken.equals( "nand" ) )
			
			moduleComponent_NAND( tokens );
			
		else if ( currentToken.equals( "nor" ) )
			
			moduleComponent_NOR( tokens );
		
		else if ( currentToken.equals( "xor" ) )
			
			moduleComponent_XOR( tokens );
			
		else if ( currentToken.equals( "xnor" ) )
			
			moduleComponent_XNOR( tokens );
			
		else if ( currentToken.equals( "dreg" ) )
			
			moduleComponent_FF_D( tokens );
		
		else if ( currentToken.equals( "notif1" ) )
			
			moduleComponent_INV3STATE( tokens );
			
		else
			
			moduleComponent_CUSTOM( tokens, currentToken );
	}
	
	private void moduleComponent_AND( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_AND c_AND = new CircuitComponent_AND( currentToken );
		
		currentComponent.addComponent( c_AND );
		
		// System.out.print( "\tAND Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			// System.out.println( "DEBUG - Adding wire: component = " + c_AND.getName() + ", gateIndex = " + componentGateIndex + ", wire = " + currentToken );
		
			currentComponent.addComponentWire( c_AND, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
		
			
		// System.out.println();
	}
	
	private void moduleComponent_OR( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_OR c_OR = new CircuitComponent_OR( currentToken );
		
		currentComponent.addComponent( c_OR );
		
		// System.out.print( "\tOR Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_OR, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
			
			
		// System.out.println();
	}
	
	private void moduleComponent_NOT( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_NOT c_NOT = new CircuitComponent_NOT( currentToken );
		
		currentComponent.addComponent( c_NOT );
		
		// System.out.print( "\tNOT Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_NOT, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}		
		
		
		tokens.next();	// Punto e virgola
		
					
		// System.out.println();
	}
	
	private void moduleComponent_NAND( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_NAND c_NAND = new CircuitComponent_NAND( currentToken );
		
		currentComponent.addComponent( c_NAND );
		
		// System.out.print( "\tNAND Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_NAND, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
		
			
		// System.out.println();
	}
	
	private void moduleComponent_NOR( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_NOR c_NOR = new CircuitComponent_NOR( currentToken );
		
		currentComponent.addComponent( c_NOR );
		
		// System.out.print( "\tNOR Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_NOR, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
		
			
		// System.out.println();
	}
	
	private void moduleComponent_XOR( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_XOR c_XOR = new CircuitComponent_XOR( currentToken );
		
		currentComponent.addComponent( c_XOR );
		
		// System.out.print( "\tXOR Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_XOR, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
		
			
		// System.out.println();
	}
	
	private void moduleComponent_XNOR( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_XNOR c_XNOR = new CircuitComponent_XNOR( currentToken );
		
		currentComponent.addComponent( c_XNOR );
		
		// System.out.print( "\tXNOR Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_XNOR, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
			
			
		// System.out.println();
	}
	
	private void moduleComponent_FF_D( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_FF_D c_FF_D = new CircuitComponent_FF_D( currentToken );
		
		// FUNZIONA SOLO PERCHE' NEL FILE VERILOG VIENE PRIMA
		// DICHIARATO IL MODULO DREG, E POI IL MODULO PRINCIPALE.
		
		copyGates( c_FF_D, "dreg" );
		
		currentComponent.addComponent( c_FF_D );
		
		// System.out.print( "\tFF_D Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_FF_D, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
		
			
		// System.out.println();
	}
	
	private void moduleComponent_INV3STATE( Iterator tokens )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent_INV3STATE c_INV3STATE = new CircuitComponent_INV3STATE( currentToken );
				
		currentComponent.addComponent( c_INV3STATE );
		
		// System.out.print( "\tFF_D Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_INV3STATE, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
		
		tokens.next();	// Punto e virgola
		
			
		// System.out.println();
	}
	
	private void moduleComponent_CUSTOM( Iterator tokens, String type )
	{
		String currentToken = ( String )( tokens.next() );
		
		CircuitComponent c_CUSTOM = new CircuitComponent( currentToken, type );
		
		copyGates( c_CUSTOM, type );
		
		currentComponent.addComponent( c_CUSTOM );
		
		// System.out.print( "\tCUSTOM Component: Name = " + currentToken + "; Pins = " );
		
		
		tokens.next();	// Parentesi aperta
		
		
		currentToken = ( String )( tokens.next() );
		
		int componentGateIndex = 0;
				
		while ( !currentToken.equals( ")" ) )
		{
			// System.out.print( currentToken + " " );
		
			currentComponent.addComponentWire( c_CUSTOM, componentGateIndex, currentToken );
			
			currentToken = ( String )( tokens.next() );
			
			componentGateIndex++;
		}	
		
			
		tokens.next();	// Punto e virgola	
		
			
		// System.out.println();
	}
	
	private void copyGates( CircuitComponent c, String type )
	{
		for ( int i = 0; i < components.size(); i++ )
		{
			CircuitComponent module = ( CircuitComponent )( components.get( i ) );
			
			if ( module.getName().equals( type ) )
			{				
				for ( int j = 0; j < module.getGateCount(); j++ )
					
					c.addGate( module.getGate( j ) );
					
				for ( int j = 0; j < module.getInputCount(); j++ )
					
					c.addInput( module.getInput( j ) );
					
				for ( int j = 0; j < module.getOutputCount(); j++ )
					
					c.addOutput( module.getOutput( j ) );
			}				
		}
			
	}
	
	private String readFileContent( String fileName )
	{
		try
		{
			String fileContent = "";
				
			FileReader fileReader = new FileReader( fileName );
			
			
			char currentChar = ( char )( fileReader.read() );
						
			while ( currentChar != 65535 )
			{
				fileContent += currentChar;
				
				currentChar = ( char )( fileReader.read() );
			}
						
			
			return fileContent;
		}
		
		catch ( FileNotFoundException e )
		{
			System.out.println( "ATTENTION! File not found: '" + fileName + "'" );
			
			return null;
		}
		
		catch ( IOException e )
		{
			System.out.println( "ATTENTION! I/O exception occurred while reading file '" + fileName + "'" );
		
			return null;
		}
	}
	
	private Iterator buildTokenIterator( String fileContent )
	{
		StringTokenizer tokenizer = new StringTokenizer( fileContent, " \n\t\r,;()", true );
		
		ArrayList tokens = new ArrayList();
		
		
		while ( tokenizer.hasMoreTokens() )
		{
			String currentToken = tokenizer.nextToken();
			
			if ( !currentToken.equals( " " ) && !currentToken.equals( "\n" ) && !currentToken.equals( "\t" ) && !currentToken.equals( "\r" ) && !currentToken.equals( "," ) )
				
				tokens.add( currentToken );
		}
		
		
		return tokens.iterator();
	}
}