package parser;

import circuit.CircuitComponent;
import circuit.CircuitLink;

import java.io.FileNotFoundException;

import java.util.Iterator;
import java.util.TreeSet;

public class VerilogToSMVConverter
{
	// --- COMPONENTI PRIMITIVI ---
	private SMVModule moduleDREG;
	
	private SMVModule moduleAND;
	
	private SMVModule moduleOR;
	
	private SMVModule moduleNOT;
	
	private SMVModule moduleNAND;
	
	private SMVModule moduleNOR;
	
	private SMVModule moduleXOR;
	
	private SMVModule moduleXNOR;
	
	private SMVModule moduleINV3STATE;
	// ----------------------------	
	
		
	public VerilogToSMVConverter( String inputFileName, String outputFileName, LTLSpecification specification ) throws FileNotFoundException
	{
		initializePrimitiveModules();
		
		VerilogParser parser = new VerilogParser( inputFileName );
		
		SMVBuilder builder = new SMVBuilder();
		
		builder.setLTLSpecification( specification );
		
		
		for ( int i = 0; i < parser.getComponentCount(); i++ )
		{
			CircuitComponent currentComponent = parser.getComponent( i );
			
			// Se il componente attuale  un flip-flop D, saltiamo al prossimo componente,
			// dal momento che esso  un componente primitivo, per il quale abbiamo gi una
			// esplicita descrizione (tale descrizione  assente nel file Verilog: vi sono
			// solo gli input e gli output).
			// NOTA: il controllo avviene sul nome, ma potrebbe indifferentemente avvenire
			// sul tipo (per i componenti dichiarati come 'module' nel file Verilog, infatti,
			// il nostro parser ne fa coincidere il nome con il tipo).
			
			if ( currentComponent.getName().equals( "dreg" ) )
			
				continue;
			
			// Creazione del modulo, con lo stesso nome del componente presente nel file Verilog
			// (stiamo parlando unicamente di quei componenti dichiarati come 'module' all'interno
			// del file Verilog).
			
			SMVModule currentModule = new SMVModule( currentComponent.getName() );
			
			// --- AGGIUNTA DELLE VARIABILI DEL MODULO ---

			// Cominciamo inserendo gli input dell'attuale componente. E' chiaro, infatti, che gli
			// input fanno parte delle variabili del relativo modulo di NuSMV.
			
			String inputsConcatenation = "";	// Concatenazione dei bit di input.
			
			String reverseInputsConcatenation = "";	// Concatenazione inversa dei bit di input.
			
			for ( int j = 0; j < currentComponent.getInputCount(); j++ )
			{
				// E' evidente che gli ingressi del componente sono delle variabili di tipo booleano.
								
				SMVVariable currentVariable = new SMVVariable( currentComponent.getInput( j ), "boolean" );
			
				currentModule.addVariable( currentVariable );
				
				// Prepariamo la concatenazione dei bit di input. La utilizzeremo pi avanti in una opportuna definizione.
				
				inputsConcatenation += "word1( " + currentComponent.getInput( j ) + " )" + ( ( j == currentComponent.getInputCount() - 1 ) ? ( "" ) : ( " :: " ) );
			
				reverseInputsConcatenation = ( ( j == currentComponent.getInputCount() - 1 ) ? ( "" ) : ( " :: " ) ) + "word1( " + currentComponent.getInput( j ) + " )" + reverseInputsConcatenation;
			}	
						
			if ( currentComponent.getInputCount() > 1 )
			{
				currentModule.addDefinition( new SMVDefinition( "INPUTS", inputsConcatenation ) );
				
				currentModule.addDefinition( new SMVDefinition( "REVERSE_INPUTS", reverseInputsConcatenation ) );
			}				
			
			// Proseguiamo inserendo gli output dell'attuale componente. E' chiaro, infatti, che anche 
			// gli output fanno parte delle variabili del relativo modulo di NuSMV.
			
			String outputsConcatenation = "";	// Concatenazione dei bit di input.
			
			String reverseOutputsConcatenation = "";	// Concatenazione inversa dei bit di output.
			
			for ( int j = 0; j < currentComponent.getOutputCount(); j++ )
			{
				// E' evidente che le uscite del componente sono delle variabili di tipo booleano.
				
				SMVVariable currentVariable = new SMVVariable( currentComponent.getOutput( j ), "boolean" );
			
				currentModule.addVariable( currentVariable );
				
				outputsConcatenation += "word1( " + currentComponent.getOutput( j ) + " )" + ( ( j == currentComponent.getOutputCount() - 1 ) ? ( "" ) : ( " :: " ) );
			
				reverseOutputsConcatenation = ( ( j == currentComponent.getOutputCount() - 1 ) ? ( "" ) : ( " :: " ) ) + "word1( " + currentComponent.getOutput( j ) + " )" + reverseOutputsConcatenation;
			}
			
			if ( currentComponent.getOutputCount() > 1 )
			{
				currentModule.addDefinition( new SMVDefinition( "OUTPUTS", outputsConcatenation ) );
			
				currentModule.addDefinition( new SMVDefinition( "REVERSE_OUTPUTS", reverseOutputsConcatenation ) );
			}				
			
			// Concludiamo inserendo i componenti utilizzati dall'attuale componente: anche essi fanno
			// parte delle variabili.			
			for ( int j = 0; j < currentComponent.getComponentCount(); j++ )
			{
				// Aggiunta della variabile corrispondente al componente j, utilizzato dal componente i.
				
				CircuitComponent currentSubComponent = currentComponent.getComponent( j );
				
				SMVVariable currentVariable = new SMVVariable( currentSubComponent.getName(), currentSubComponent.getType() );
			
				currentModule.addVariable( currentVariable );
				
				
				// Controllo: se vengono usati componenti primitivi, inseriamo il relativo modulo.
				// Se un tale modulo  gi stato inserito, esso verr aggiunto comunque, ma comparir
				// una volta sola: ci grazie al metodo equals() ridefinito nella classe SMVModule
				// (due moduli sono uguali se e solo se hanno lo stesso nome).
				
				if ( currentSubComponent.getType().equals( "dreg" ) )
					
					builder.addModule( moduleDREG );
					
				else if ( currentSubComponent.getType().equals( "and" ) )
					
					builder.addModule( moduleAND );
					
				else if ( currentSubComponent.getType().equals( "or" ) )
					
					builder.addModule( moduleOR );
					
				else if ( currentSubComponent.getType().equals( "not" ) )
					
					builder.addModule( moduleNOT );
					
				else if ( currentSubComponent.getType().equals( "nand" ) )
				
					builder.addModule( moduleNAND );
				
				else if ( currentSubComponent.getType().equals( "nor" ) )
				
					builder.addModule( moduleNOR );
				
				else if ( currentSubComponent.getType().equals( "exor" ) )
				
					builder.addModule( moduleXOR );
				
				else if ( currentSubComponent.getType().equals( "exnor" ) )	
			
					builder.addModule( moduleXNOR );
					
				else if ( currentSubComponent.getType().equals( "inv3state" ) )	
				{
					builder.addModule( moduleINV3STATE );
					
					// --- VECCHIA SOLUZIONE, NON PIU' UTILIZZATA PERCHE' NON FUNZIONANTE ---
					// String v = currentVariable.getName();
					
					// specification.addEntailer( new LTLSpecification( "G ( ( " + v + ".enable -> " + v + ".output = !" + v + ".input ) & ( !" + v + ".enable -> " + v + ".output = z ) )" ) );
					// ----------------------------------------------------------------------
				}					
			}				
			
			// -------------------------------------------
			
			// --- AGGIUNTA DELLE ASSEGNAZIONI DEL MODULO ---
			
			// Tramite le assegnazioni, possiamo modellare i collegamenti tra i componenti
			// facenti parte di un componente pi complesso.
			
			for ( int j = 0; j < currentComponent.getLinkCount(); j++ )
			{
				CircuitLink currentLink = currentComponent.getLink( j );
				
				CircuitComponent c_IN = currentLink.getComponentIn();
				
				CircuitComponent c_OUT = currentLink.getComponentOut();
				
				String leftSide, rightSide;
				
				SMVAssignment currentAssignment;
				
				if ( c_IN != null & c_OUT != null )
				{
					// Se il componente che gioca il ruolo di output (cio fornitore del segnale)
					//  un inverter 3-state...
					
					if ( c_OUT.getType().equals( "inv3state" ) )
					{
						// ...non possiamo aggiungere una assegnazione. Infatti, NuSMV
						// troverebbe un errore di sintassi: non possiamo associare un valore (output) che appartiene
						// al dominio { TRUE, FALSE, z } ad un valore (input) che appartiene al dominio { TRUE, FALSE }. 
						// "Dobbiamo aggiungere un entailer alla specifica LTL."
						// La soluzione finale non  pi di aggiungere un entailer, ma di utilizzare
						// gli invarianti.
						
						String v1 = c_OUT.getName();	// Nome dell'inverter 3-state.
						
						String v2 = c_IN.getName();		// Nome del componente che gioca il ruolo di input (cio utilizzatore del segnale)
					
						String in = c_IN.getInput( currentLink.getIndexIn() );	// Nome dello specifico ingresso di input.
					
						// --- VECCHIA SOLUZIONE, NON PIU' UTILIZZATA PERCHE' NON FUNZIONANTE ---
						// Entailer poco conciso. Commentato.
					
						// specification.addEntailer( new LTLSpecification( "G ( " + v1 + ".output != z -> ( ( " + v1 + ".output = TRUE -> " + v2 + "." + in + " = TRUE ) & ( " + v1 + ".output = FALSE -> " + v2 + "." + in + " = FALSE ) ) )" ) );
					
						// Entailer pi conciso. Migliore.
					
						// specification.addEntailer( new LTLSpecification( "G ( " + v1 + ".output != z -> " + v2 + ".in = " + v1 + ".output )" ) );
						// ----------------------------------------------------------------------
						
						// Aggiunta degli invarianti.
						
						currentModule.addInvariant( new SMVInvariant( v1 + ".output = TRUE -> " + v2 + "." + in + " = TRUE" ) );
						
						currentModule.addInvariant( new SMVInvariant( v1 + ".output = FALSE -> " + v2 + "." + in + " = FALSE" ) );
					}
					
					else
					{
						// ...altrimenti, se non si tratta di un inverter 3-state, possiamo
						// tranquillamente adoperare una assegnazione.
						
						leftSide = c_IN.getName() + "." + c_IN.getInput( currentLink.getIndexIn() );
					
						rightSide = c_OUT.getName() + "." + c_OUT.getOutput( currentLink.getIndexOut() );
					
						currentAssignment = new SMVAssignment( leftSide, rightSide );
					
						currentModule.addAssignment( currentAssignment );	
					}
				}
				
				// INPUTS DEL COMPONENTE IN ESAME
					
				if ( currentComponent.isInput( currentLink.getWire() ) )
				{
					leftSide = c_IN.getName() + "." + c_IN.getInput( currentLink.getIndexIn() );
					
					rightSide = currentLink.getWire();
					
					currentAssignment = new SMVAssignment( leftSide, rightSide );
					
					currentModule.addAssignment( currentAssignment );	
				}	
				
				// OUTPUTS DEL COMPONENTE IN ESAME
					
				if ( currentComponent.isOutput( currentLink.getWire() ) ) 	
				{
					// Stesso procedimento usato poco sopra: se il componente che alimenta
					// questo output  un inverter 3-state, dobbiamo... ( "aggiungere un entailer
					// alla specifica LTL (non possiamo utilizzare una assegnazione, essendo
					// presente, nel dominio, anche la variabile z).", vecchia soluzione )
					// ...aggiungere un invariante ( nuova soluzione ).
					
					if ( c_OUT.getType().equals( "inv3state" ) )
					{
						String v = c_OUT.getName();		// Nome dell'inverter 3-state.
						
						String out = currentLink.getWire();		// Nome di questo output.
						
						// --- VECCHIA SOLUZIONE, NON PIU' UTILIZZATA PERCHE' NON FUNZIONANTE ---
						// specification.addEntailer( new LTLSpecification( "G ( " + v + ".output != z -> " + out + " = " + v + ".output )" ) );
						// ----------------------------------------------------------------------
						
						currentModule.addInvariant( new SMVInvariant( v + ".output = TRUE -> " + out + " = TRUE" ) );
						
						currentModule.addInvariant( new SMVInvariant( v + ".output = FALSE -> " + out + " = FALSE" ) );
					}
					
					else
					{				
						leftSide = currentLink.getWire();
						
						rightSide = c_OUT.getName() + "." + c_OUT.getOutput( currentLink.getIndexOut() );
											
						currentAssignment = new SMVAssignment( leftSide, rightSide );
						
						currentModule.addAssignment( currentAssignment );
					}	
				}
				
				// COLLEGAMENTI A 1 (TENSIONE NON NULLA)
				
				if ( currentLink.getWire().equals( "vdd" ) )
				{
					// 1 Logico: tensione non nulla.
					
					leftSide = c_IN.getName() + "." + c_IN.getInput( currentLink.getIndexIn() );
						
					rightSide = "TRUE";
										
					currentAssignment = new SMVAssignment( leftSide, rightSide );
					
					currentModule.addAssignment( currentAssignment );
				}
				
				// COLLEGAMENTI A 0 (TERRA)
				
				if ( currentLink.getWire().equals( "vss" ) )
				{
					// 0 Logico: tensione nulla (terra).
					
					leftSide = c_IN.getName() + "." + c_IN.getInput( currentLink.getIndexIn() );
						
					rightSide = "FALSE";
										
					currentAssignment = new SMVAssignment( leftSide, rightSide );
					
					currentModule.addAssignment( currentAssignment );
				}
			}
			
			// ----------------------------------------------
			
			builder.addModule( currentModule );
		}
		
		builder.saveToFile( outputFileName );			
	}

	private void initializePrimitiveModules()
	{
		moduleDREG = new SMVModule( "dreg" );
		
		moduleDREG.addVariable( new SMVVariable( "d", "boolean" ) );
				
		moduleDREG.addVariable( new SMVVariable( "q", "boolean" ) );	
	
		moduleDREG.addDefinition( new SMVDefinition( "nq", "!q" ) );
	
		moduleDREG.addTransition( new SMVTransition( "d = 1", "next(q) = 1" ) );
		
		moduleDREG.addTransition( new SMVTransition( "d = 0", "next(q) = 0" ) );
		
		
		moduleAND = new SMVModule( "and" );
		
		moduleAND.addVariable( new SMVVariable( "in_1", "boolean" ) );
		
		moduleAND.addVariable( new SMVVariable( "in_2", "boolean" ) );
		
		moduleAND.addDefinition( new SMVDefinition( "out", "in_1 & in_2" ) );
		
		
		moduleOR = new SMVModule( "or" );
		
		moduleOR.addVariable( new SMVVariable( "in_1", "boolean" ) );
		
		moduleOR.addVariable( new SMVVariable( "in_2", "boolean" ) );
		
		moduleOR.addDefinition( new SMVDefinition( "out", "in_1 | in_2" ) );
		
		
		moduleNOT = new SMVModule( "not" );
		
		moduleNOT.addVariable( new SMVVariable( "input", "boolean" ) );
				
		moduleNOT.addDefinition( new SMVDefinition( "output", "!input" ) );
		
		
		moduleNAND = new SMVModule( "nand" );
		
		moduleNAND.addVariable( new SMVVariable( "in_1", "boolean" ) );
		
		moduleNAND.addVariable( new SMVVariable( "in_2", "boolean" ) );
		
		moduleNAND.addDefinition( new SMVDefinition( "out", "!( in_1 & in_2 )" ) );
		
		
		moduleNOR = new SMVModule( "nor" );
		
		moduleNOR.addVariable( new SMVVariable( "in_1", "boolean" ) );
		
		moduleNOR.addVariable( new SMVVariable( "in_2", "boolean" ) );
		
		moduleNOR.addDefinition( new SMVDefinition( "out", "!( in_1 | in_2 )" ) );
		
		
		moduleXOR = new SMVModule( "exor" );
		
		moduleXOR.addVariable( new SMVVariable( "in_1", "boolean" ) );
		
		moduleXOR.addVariable( new SMVVariable( "in_2", "boolean" ) );
		
		moduleXOR.addDefinition( new SMVDefinition( "out", "( !in_1 & in_2 ) | ( in_1 & !in_2 )" ) );
		
		
		moduleXNOR = new SMVModule( "exnor" );
		
		moduleXNOR.addVariable( new SMVVariable( "in_1", "boolean" ) );
		
		moduleXNOR.addVariable( new SMVVariable( "in_2", "boolean" ) );
		
		moduleXNOR.addDefinition( new SMVDefinition( "out", "( in_1 | !in_2 ) & ( !in_1 | in_2 )" ) );
		
		
		moduleINV3STATE = new SMVModule( "inv3state" );
		
		moduleINV3STATE.addVariable( new SMVVariable( "input", "boolean" ) );
		
		moduleINV3STATE.addVariable( new SMVVariable( "enable", "boolean" ) );
		
		// La Z (alta impedenza) deve essere minuscola, altrimenti NuSMV d problemi di sintassi.
		
		moduleINV3STATE.addVariable( new SMVVariable( "output", "{ TRUE, FALSE, z }" ) );
		
		moduleINV3STATE.addInvariant( new SMVInvariant( "enable -> ( ( input = TRUE -> output = FALSE ) & ( input = FALSE -> output = TRUE ) )" ) );
	
		moduleINV3STATE.addInvariant( new SMVInvariant( "!enable -> output = z" ) );
	}
}